/*******************************************************************************
***  This file is part of Dux Rei Nummariae.                                 ***
***                                                                          ***
***  Copyright (C) 2017, 2018, 2019, 2020, 2021, 2022, 2023                  ***
***  John Schneiderman <Licencing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Affero General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Affero General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Affero General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#ifndef DRN_FOUNDATION_QTMEMORY_HPP_
#define DRN_FOUNDATION_QTMEMORY_HPP_

#include <QMetaObject>
#include <QObject>

#include <cassert>
#include <memory>
#include <type_traits>

#include <foundation/ObserverPtr.hpp>


class QListWidgetItem;
class QTableWidgetItem;
class QTreeWidgetItem;

namespace drn::foundation::internal
{

/**
 * @brief Represents the idea of a Qt object that derives from ::QObject.
 *
 * @tparam ObjectType The type to inspect if it's derived from ::QObject.
 */
template<typename ObjectType>
concept QObjectBase = std::is_base_of_v<::QObject, ObjectType>;

/**
 * @brief Represents the idea of a Qt type whose memory is handled by the parent-child pattern.
 *
 * @tparam ObjectType The type to inspect if it's manageable by Qt.
 */
template<typename ObjectType>
concept QtParentableObject = QObjectBase<ObjectType>
		|| std::is_same_v<::QTreeWidgetItem, ObjectType>;


/**
 * @brief Represents the idea of any Qt public type.
 *
 * @tparam ObjectType The type to inspect if it's created by the Qt SDK.
 */
template<typename ObjectType>
concept QtType =
	QObjectBase<ObjectType>
		|| std::is_same_v<::QListWidgetItem, ObjectType>
		|| std::is_same_v<::QTableWidgetItem, ObjectType>
		|| std::is_same_v<::QTreeWidgetItem, ObjectType>;

template<typename QtPtrType>
requires requires (QtPtrType p)
{
	typename QtPtrType::element_type;
	p == nullptr;
	{ p.get() } -> std::same_as<typename QtPtrType::element_type*>;
	p = nullptr;
}
QMetaObject::Connection createDestroyableConnection(QtPtrType& self);

}
namespace drn::foundation
{

template<typename>
class UniqueQtPtr;

template<typename ObjectType>
class QtPtr : public ObserverPtr<ObjectType>
{
	QMetaObject::Connection connection_;

public:
	QtPtr(ObjectType* const observe = nullptr)
			requires internal::QtParentableObject<ObjectType>;
	QtPtr(const UniqueQtPtr<ObjectType>& observe);
	QtPtr(const QtPtr&) = default;
	QtPtr(QtPtr&&) noexcept = default;
	QtPtr& operator=(const QtPtr&) = default;
	QtPtr& operator=(QtPtr&&) noexcept = default;
	~QtPtr();
	using ObserverPtr<ObjectType>::get;
	using ObserverPtr<ObjectType>::operator*;
	using ObserverPtr<ObjectType>::operator->;
	using ObserverPtr<ObjectType>::operator==;
	using ObserverPtr<ObjectType>::operator!=;
	using ObserverPtr<ObjectType>::operator<;
};

/**
 * @brief A representation of a Qt object that is not handled by the usual Qt memory pattern. When
 * the Qt object derives from a QObject its deletion is handled on the event loop, else it's deleted
 * immediately upon destruction.
 *
 * @tparam ObjectType The type of the object stored.
 */
template<typename ObjectType>
class UniqueQtPtr final : std::unique_ptr<ObjectType, void(*)(ObjectType*)>
{
public:
	explicit UniqueQtPtr(ObjectType* const obj) requires internal::QtType<ObjectType>;
	using std::unique_ptr<ObjectType, void(*)(ObjectType*)>::operator*;
	using std::unique_ptr<ObjectType, void(*)(ObjectType*)>::operator->;
	using std::unique_ptr<ObjectType, void(*)(ObjectType*)>::get;
	using std::unique_ptr<ObjectType, void(*)(ObjectType*)>::release;
	using std::unique_ptr<ObjectType, void(*)(ObjectType*)>::swap;
	bool operator==(std::nullptr_t) const noexcept;
	bool operator!=(std::nullptr_t) const noexcept;
};

/**
 * @brief Creates an object using the Qt memory pattern. Guarantees the object is of a Qt type that
 * can participate in the Qt memory management and is owned by a parent. Additionally, guarantees
 * the type is a Qt type.
 *
 * @tparam ObjectType The type of the object to create.
 * @tparam ObjectArgTypes The argument types for the object constructor to invoke.
 *
 * @param objectArgs The values of to use in the object creation.
 *
 * @return The Qt object that is created on the heap.
 */
template<typename ObjectType, typename ... ObjectArgTypes>
requires internal::QtParentableObject<ObjectType>
foundation::QtPtr<ObjectType> makeQtPtr(ObjectArgTypes&& ... objectArgs);

/**
 * @brief Creates an object using the Qt memory pattern. Guarantees the object is of a Qt type that
 * can participate in the Qt memory management and is owned by a parent.
 *
 * @tparam ObjectType The type of the object to create.
 * @tparam BaseType The base type of the object being created.
 * @tparam ObjectArgTypes The argument types for the object constructor to invoke.
 *
 * @param objectArgs The values of to use in the object creation.
 *
 * @return The Qt object that is created on the heap.
 */
template<typename ObjectType, typename BaseType, typename ... ObjectArgTypes>
requires internal::QtParentableObject<ObjectType>
foundation::QtPtr<BaseType> makeBaseQtPtr(ObjectArgTypes&& ... objectArgs);

/**
 * @brief Creates a Qt object that is not handled by the Qt memory pattern.
 *
 * @tparam ObjectType The type of the object to create.
 * @tparam ObjectArgTypes The argument types for the object constructor to invoke.
 *
 * @param objectArgs The values of to use in the object creation.
 *
 * @return The Qt object that is created on the heap.
 */
template<typename ObjectType, typename ... ObjectArgTypes>
foundation::UniqueQtPtr<ObjectType> makeUniqueQtPtr(ObjectArgTypes&& ... objectArgs);

/**
 * @brief Creates a Qt object that is not handled by the Qt memory pattern.
 *
 * @tparam ObjectType The type of the object to create.
 * @tparam BaseType The base type of the object being created.
 * @tparam ObjectArgTypes The argument types for the object constructor to invoke.
 *
 * @param objectArgs The values of to use in the object creation.
 *
 * @return The Qt object that is created on the heap.
 */
template<typename ObjectType, typename BaseType, typename ... ObjectArgTypes>
foundation::UniqueQtPtr<ObjectType> makeUniqueBaseQtPtr(ObjectArgTypes&& ... objectArgs);

template<typename ObservedType, typename ObservedPtrType>
bool operator==(const ObservedType* lhs, const QtPtr<ObservedPtrType>& rhs);

}

template<typename QtPtrType>
requires requires (QtPtrType p)
{
	typename QtPtrType::element_type;
	p == nullptr;
	{ p.get() } -> std::same_as<typename QtPtrType::element_type*>;
	p = nullptr;
}
QMetaObject::Connection drn::foundation::internal::createDestroyableConnection(QtPtrType& self)
{
	if (self.get() == nullptr)
		return {};

	if constexpr (internal::QObjectBase<typename QtPtrType::element_type*>)
	{
		return ::QObject::connect(
			self.get(),
			&::QObject::destroyed,
			[&self] (auto obj) mutable
			{
				assert(
					self.get() == obj
					&& "Should be the same object being destroyed."
				);
				self = nullptr;
			}
		);
	}
	return {};
}

template<typename ObjectType>
drn::foundation::QtPtr<ObjectType>::QtPtr(ObjectType* const observe)
		requires drn::foundation::internal::QtParentableObject<ObjectType> :
	ObserverPtr<ObjectType>{observe},
	connection_{internal::createDestroyableConnection(*this)}
{}

template<typename ObjectType>
drn::foundation::QtPtr<ObjectType>::QtPtr(const UniqueQtPtr<ObjectType>& observe) :
	drn::foundation::QtPtr<ObjectType>{observe.get()}
{}

template<typename ObjectType>
drn::foundation::QtPtr<ObjectType>::~QtPtr()
{
	if constexpr (internal::QObjectBase<ObjectType>)
		QObject::disconnect(this->connection_);
}

template<typename ObjectType>
drn::foundation::UniqueQtPtr<ObjectType>::UniqueQtPtr(ObjectType* const obj)
		requires internal::QtType<ObjectType> :
	std::unique_ptr<ObjectType, void(*)(ObjectType*)>{
		obj,
		[] (ObjectType* o)
		{
			if constexpr (internal::QObjectBase<ObjectType>)
			{
				if (o != nullptr)
					o->deleteLater();
			}
			else
			{
				if (o != nullptr)
					delete o;
			}
		}
	}
{
	if constexpr (internal::QObjectBase<ObjectType>)
	{
		if (obj != nullptr)
			assert(obj->parent() == nullptr && "All unique Qt objects must not have a parent.");
	}
}

template<typename ObjectType>
bool drn::foundation::UniqueQtPtr<ObjectType>::operator==(std::nullptr_t) const noexcept
{
	return this->get() == nullptr;
}

template<typename ObjectType>
bool drn::foundation::UniqueQtPtr<ObjectType>::operator!=(std::nullptr_t) const noexcept
{
	return ! (*this == nullptr);
}

template<typename ObjectType, typename ... ObjectArgTypes>
requires drn::foundation::internal::QtParentableObject<ObjectType>
drn::foundation::QtPtr<ObjectType> drn::foundation::makeQtPtr(ObjectArgTypes&& ... objectArgs)
{
	return makeBaseQtPtr<ObjectType, ObjectType>(std::forward<ObjectArgTypes>(objectArgs)...);
}

template<typename ObjectType, typename BaseType, typename ... ObjectArgTypes>
requires drn::foundation::internal::QtParentableObject<ObjectType>
drn::foundation::QtPtr<BaseType> drn::foundation::makeBaseQtPtr(ObjectArgTypes&& ... objectArgs)
{
	QtPtr<BaseType> obj{new ObjectType{std::forward<ObjectArgTypes>(objectArgs)...}};

	// BUG: Workaround for the QTreeWidgetItem parent call returning null even though the supplied parent is not null with Qt 5.15.
	if constexpr (std::is_same_v<::QObject, BaseType>)
		assert(obj->parent() != nullptr && "All Qt objects must have a parent.");
	return obj;
}

template<typename ObjectType, typename ... ObjectArgTypes>
drn::foundation::UniqueQtPtr<ObjectType> drn::foundation::makeUniqueQtPtr(
	ObjectArgTypes&& ... objectArgs
)
{
	return makeUniqueBaseQtPtr<ObjectType, ObjectType>(std::forward<ObjectArgTypes>(objectArgs)...);
}

template<typename ObjectType, typename BaseType, typename ... ObjectArgTypes>
drn::foundation::UniqueQtPtr<ObjectType> drn::foundation::makeUniqueBaseQtPtr(
	ObjectArgTypes&& ... objectArgs
)
{
	UniqueQtPtr<ObjectType> ptr{new BaseType{std::forward<ObjectArgTypes>(objectArgs)...}};
	return ptr;
}

template<typename ObservedType, typename ObservedPtrType>
bool drn::foundation::operator==(
	const ObservedType* lhs,
	const drn::foundation::QtPtr<ObservedPtrType>& rhs
)
{
	return lhs == static_cast<ObserverPtr<ObservedPtrType>>(rhs);
}

#endif
