/*******************************************************************************
***  This file is part of Dux Rei Nummariae.                                 ***
***                                                                          ***
***  Copyright (C) 2017, 2018, 2019, 2020, 2021, 2022, 2023                  ***
***  John Schneiderman <Licencing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Affero General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Affero General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Affero General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#ifndef DRN_STORAGE_BUDGETSTORAGE_HPP_
#define DRN_STORAGE_BUDGETSTORAGE_HPP_

#include <map>
#include <tuple>


namespace drn
{
namespace accounting
{

class AccountCode;
class AccountNumber;
class GeneralLedger;

}
namespace banking
{

enum class SupportedAccountTypes;
struct BankMap;
class BankName;

}
namespace budgeting
{

class Budget;
class BudgetSource;

}
namespace foundation
{

template<typename, typename>
class TypeIndexMap;

}
namespace surveying
{

class ReconciledSurveys;

}
namespace storage
{

/**
 * @brief The manager interface for interacting with a stored budget.
 */
struct BudgetStorage
{
	BudgetStorage() = default;
	BudgetStorage(const BudgetStorage&) = default;
	BudgetStorage(BudgetStorage&&) = default;
	virtual ~BudgetStorage() = default;
	BudgetStorage& operator=(const BudgetStorage&) = default;
	BudgetStorage& operator=(BudgetStorage&&) = default;
	/**
	 * @brief When implemented, provides an accessor to the name of the budget stored, e.g. when
	 * stored on a file-system this would be the file-name.
	 */
	virtual const ::QString& name() const noexcept = 0;
	/**
	 * @brief When implemented, changes the storage location for storing or loading a budget.
	 *
	 * @exception StorageError When the path is invalid.
	 *
	 * @param location The path to the location of the budget.
	 */
	virtual void changeLocation(const ::QString& location) = 0;
	/**
	 * @brief When implemented, loads a budget file from a storage location.
	 *
	 * @exception StorageError When there's an issue loading the budget.
	 *
	 * @return The the general ledger (first), fully loaded budget (second), the mapping between
	 * each budget item and its ledger account number (third), the financial institutions
	 * (fourth), and  the financial institution associations in the general ledger (fifth)
	 */
	virtual std::tuple<
		accounting::GeneralLedger,
		budgeting::Budget,
		foundation::TypeIndexMap<budgeting::BudgetSource, accounting::AccountCode>,
		banking::BankMap,
		std::map<
			banking::BankName,
			std::map<accounting::AccountNumber, banking::SupportedAccountTypes>
		>,
		surveying::ReconciledSurveys
	> load() const = 0;
	/**
	 * @brief When implemented, stores all budget information to a location.
	 *
	 * @exception StorageError When the budget cannot be stored.
	 *
	 * @param[in] gl The general ledger whose contents are to be stored in the file.
	 * @param[in] b The budget whose contents are to be stored in the file.
	 * @param[in] budgetCodes The mapping between each budget item and its ledger account number.
	 * @param[in] banks The financial institutions whose contents are to be stored in the file.
	 * @param[in] bankAccountTypes The associations between the financial institutions and the
	 * general ledger whose contents are to be stored in the file.
	 */
	virtual void store(
		const accounting::GeneralLedger& gl,
		const budgeting::Budget& b,
		const foundation::TypeIndexMap<
			budgeting::BudgetSource,
			accounting::AccountCode
		>& budgetCodes,
		const banking::BankMap& banks,
		const std::map<
			banking::BankName,
			std::map<accounting::AccountNumber, banking::SupportedAccountTypes>
		>& bankAccountTypes,
		const surveying::ReconciledSurveys& surveys
	) const = 0;
};

}}

#endif
