/*******************************************************************************
***  This file is part of Dux Rei Nummariae.                                 ***
***                                                                          ***
***  Copyright (C) 2017, 2018, 2019, 2020, 2021, 2022, 2023                  ***
***  John Schneiderman <Licencing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Affero General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Affero General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Affero General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#ifndef DRN_SURVEYING_RECONCILEDSURVEY_H_
#define DRN_SURVEYING_RECONCILEDSURVEY_H_

#include <pecunia/Money.h>
#include <QString>

#include <iosfwd>
#include <map>

#include <budgeting/BudgetItemIdentifier.h>
#include <banking/BankAccount.h>
#include <banking/Reconciliation.h>
#include <foundation/Optional.hpp>

#include "surveying_export.h"


class QDate;

namespace drn
{
namespace accounting
{

class AccountNumber;

}
namespace surveying
{

class DRN_SURVEYING_EXPORT ReconciledSurvey
{
	banking::ReconciledBankAccount reconciled_;
	std::map<budgeting::BudgetItemIdentifier, pecunia::currency::Money> distribution_;
	foundation::Optional<::QString> checksum_;

public:
	ReconciledSurvey() = default;
	/**
	 * @brief Partial initialisation constructor for creating a new survey.
	 *
	 * @param ba The bank account that is reconciled.
	 * @param currency The currency held within the bank account.
	 */
	ReconciledSurvey(banking::BankAccount ba, const pecunia::currency::Iso4217Codes& currency);
	/**
	 * @brief Full initialisation constructor.
	 *
	 * @exception Error When a supplied checksum value does not match the calculated value.
	 *
	 * @param reconciled The bank account that was reconciled.
	 * @param distribution The distribution of the balance in the account.
	 * @param checksum The value that represents an object whose values were verified after having
	 * completed a survey. This is in lieu of verifying the balances upon object creation since
	 * exchange rates change over time preventing verification in the future.
	 */
	ReconciledSurvey(
		banking::ReconciledBankAccount reconciled,
		std::map<
			budgeting::BudgetItemIdentifier,
			pecunia::currency::Money
		> distribution,
		foundation::Optional<::QString> checksum
	);
	const banking::ReconciledBankAccount& reconciled() const noexcept;
	const std::map<
		budgeting::BudgetItemIdentifier,
		pecunia::currency::Money
	>& distribution() const noexcept;
	/**
	 * @brief The value that represents a valid object after having completed a survey.
	 */
	const foundation::Optional<::QString>& checksum() const noexcept;
	/**
	 * @brief Updates the survey with the completed values.
	 *
	 * @exception BankError When the reconciled bank balance does not match the amount of
	 * distributed unused funds.
	 *
	 * @param completedOn The date the balance was confirmed.
	 * @param reconciledBalance The balance as reported in the bank account.
	 * @param distribution The distribution of unused funds (bank balance) in the bank account.
	 *
	 * @post Upon successful survey completion, the validity value is computed and stored in the object.
	 */
	void completed(
		const ::QDate& completedOn,
		const pecunia::currency::Money& reconciledBalance,
		const std::map<budgeting::BudgetItemIdentifier, pecunia::currency::Money>& distribution
	);
	void renamed(const banking::BankName& original, const banking::BankName& newName);
	void renamed(const accounting::AccountNumber& original, const accounting::AccountCode& changed);
	void renamed(const budgeting::BudgetItemIdentifier& id, const budgeting::BudgetSource& newName);
};

class DRN_SURVEYING_EXPORT ReconciledSurveys : std::map<banking::BankAccount, ReconciledSurvey>
{
public:
	explicit ReconciledSurveys(std::map<banking::BankAccount, ReconciledSurvey> surveys);
	using std::map<banking::BankAccount, ReconciledSurvey>::map;
	using std::map<banking::BankAccount, ReconciledSurvey>::const_iterator;
	using std::map<banking::BankAccount, ReconciledSurvey>::cbegin;
	using std::map<banking::BankAccount, ReconciledSurvey>::cend;
	using std::map<banking::BankAccount, ReconciledSurvey>::begin;
	using std::map<banking::BankAccount, ReconciledSurvey>::end;
	using std::map<banking::BankAccount, ReconciledSurvey>::empty;
	using std::map<banking::BankAccount, ReconciledSurvey>::count;
	using std::map<banking::BankAccount, ReconciledSurvey>::at;
	using std::map<banking::BankAccount, ReconciledSurvey>::size;
	foundation::Optional<ReconciledSurvey> find(const banking::BankAccount& ba) const;
	foundation::Optional<ReconciledSurvey> find(
		const banking::BankName& bn,
		const accounting::AccountNumber& an
	) const;
	bool upsert(ReconciledSurvey survey);
};

DRN_SURVEYING_EXPORT bool operator==(
	const ReconciledSurvey& lhs,
	const ReconciledSurvey& rhs
);
DRN_SURVEYING_EXPORT bool operator!=(
	const ReconciledSurvey& lhs,
	const ReconciledSurvey& rhs
);
DRN_SURVEYING_EXPORT bool operator==(
	const ReconciledSurveys& lhs,
	const ReconciledSurveys& rhs
);
DRN_SURVEYING_EXPORT bool operator!=(
	const ReconciledSurveys& lhs,
	const ReconciledSurveys& rhs
);

DRN_SURVEYING_EXPORT std::ostream& operator<<(std::ostream& o, const ReconciledSurvey& rs);
DRN_SURVEYING_EXPORT std::ostream& operator<<(std::ostream& o, const ReconciledSurveys& rss);

}}

#endif
